/* EOPL syntax for typed source language */
/* Stephanie Weirich */

%token LAMBDA LPAREN RPAREN PLUS TIMES MINUS LET COMMA SEMI
%token <string>ID
%token <int> NUMBER
%token TRUE FALSE 
%token IF EQUAL AND OR NOT THEN ELSE IN LETREC SET
%token LT GT ZERO ADD1 SUB1
%token BEGIN END SET REF CELL CONTENTS SETCELL DEREF SETREF
%token CAR CDR CONS ISNULL NULL INT BOOL LBRACKET RBRACKET COLON LIST ARROW
%token PAIR FST SND LETTYPE MARK

%start main
%type <Absyn.exp> main

%%

  main: 
    exp { $1 }
  ;

  typ: 
    INT  { Absyn.IntType } 
  | BOOL { Absyn.BoolType }
  | LPAREN typs ARROW typ RPAREN { Absyn.ProcType ($2,$4) } 
  | LIST LPAREN typ RPAREN { Absyn.ListType($3) }
  | LPAREN typ TIMES typ RPAREN { Absyn.PairType ($2,$4) }
 
  typs:  { [] } 
  | typ { [ $1 ] } 
  | typ TIMES typs { $1 :: $3 }

  exp: 
    ID { Absyn.Var ($1) }
  | prim LPAREN cexps RPAREN { Absyn.Prim($1,$3) }
  | NUMBER { Absyn.LitInt ($1) }
  | TRUE   { Absyn.LitBool (true) }
  | FALSE  { Absyn.LitBool (false) }
  | IF exp THEN exp ELSE exp 
      { Absyn.If ($2,$4,$6) }
  | LET binds IN exp
      { Absyn.Let (fst $2,snd $2, $4) }
  | LAMBDA LPAREN id_tys RPAREN exp 
      { Absyn.Lambda (fst $3, snd $3, $5) }
  | LPAREN exp exps RPAREN 
      { Absyn.App ($2,$3) }
  | LETREC recbinds IN exp 
      { let (a,b,c,d,e) = $2 in Absyn.Letrec(a,b,c,d,e,$4) }		
  | BEGIN exp SEMI sexps END 
      { Absyn.Begin ($2,$4) }  
  | SET ID EQUAL exp 
      { Absyn.Varassign($2,$4) }
  | NULL
		{ Absyn.LitNull }
  ;

  id_tys:   { [],[] } 
  | typ_or_mark ID { ([$1],[$2]) }
  | typ_or_mark ID COMMA id_tys { ($1::fst $4,$2::snd $4) }
  ;

  typ_or_mark:
	 typ  { ref(Some $1) }
  | MARK { ref(None) }
  ;

  binds:   { [],[] }
  | ID EQUAL exp binds { ($1::fst $4,$3::snd $4) }
  ; 

  recbinds: { [],[],[],[],[] }
  | typ_or_mark ID LPAREN id_tys RPAREN EQUAL exp recbinds
      { let (typs, procnames, argtyps, idss, bodies) = $8 in 
	  ($1::typs, $2::procnames, fst $4::argtyps, snd $4::idss, $7::bodies) }
  ;

  prim: PLUS  { Absyn.Plus  }
  | TIMES { Absyn.Times }
  | MINUS { Absyn.Minus }
  | EQUAL { Absyn.Equal }
  | AND   { Absyn.And   }
  | OR    { Absyn.Or    }
  | NOT   { Absyn.Not   }
  | ADD1  { Absyn.Add1  }
  | SUB1  { Absyn.Sub1  }
  | LT    { Absyn.LT    }
  | GT    { Absyn.GT    }
  | ZERO  { Absyn.Zero  }
  | CONS     { Absyn.Cons }
  | CAR      { Absyn.Hd }
  | CDR      { Absyn.Tl }
  | ISNULL   { Absyn.IsNull }
  | PAIR     { Absyn.Pair }
  | FST      { Absyn.Fst }
  | SND      { Absyn.Snd }
  ;

  exps:     { [] } 
  | exp exps  { ($1 :: $2) }
  ;

  cexps:    exp  { [$1] } 
  | exp COMMA cexps  { ($1 :: $3) }
  ;

  sexps:    exp  { [$1] } 
  | exp SEMI sexps  { ($1 :: $3) }

%%
